"""
"""


#[
from __future__ import annotations

import functools as _ft
import numpy as _np

from .. import dates as _dates
from . import _functionalize
#]


__all__ = ()


class Inlay:
    """
    """
    #[

    def fill_missing(
        self,
        fill_range: Iterable[_dates.Dater],
        method: Literal["next", "previous", "nearest", "constant"],
        *args,
        **kwargs,
    ) -> None:
        """
        """
        fill_func = _METHOD_FACTORY[method]
        data = self.get_data(fill_range, )
        new_data = [
            fill_func(variant, *args, **kwargs, ).T
            for variant in data.T
        ]
        self.set_data(fill_range, new_data, )

    #]


for n in ("fill_missing", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _fill_neighbor(
    data,
    func,
    *args,
    **kwargs,
) -> _np.ndarray:
    """
    """
    index_nan = _np.isnan(data)
    if _np.all(index_nan):
        return data
    where_nan = _np.where(index_nan)[0]
    where_obs = _np.where(~index_nan)[0].astype(_np.float64, )
    if where_obs.size == 0:
        return data
    for i in where_nan:
        j = func(i, where_obs, )
        data[i] = data[int(j)] if j is not None else _np.nan
    return data


def _fill_constant(
    data,
    constant,
    /,
    *args,
    **kwargs,
) -> _np.ndarray:
    """
    """
    index_nan = _np.isnan(data)
    data[index_nan] = constant
    return data


def _next_index(i: int, where_obs: _np.ndarray, /, ) -> int:
    """
    """
    #[
    where_obs_minus_i = where_obs - i
    index_negative = where_obs_minus_i < 0
    if _np.all(index_negative):
        return None
    where_obs_minus_i[index_negative] = _np.inf
    index = _np.argmin(where_obs_minus_i)
    return where_obs[index]
    #]


def _previous_index(i: int, where_obs: _np.ndarray, /, ) -> int:
    """
    """
    #[
    i_minus_where_obs = i - where_obs
    index_negative = i_minus_where_obs < 0
    if _np.all(index_negative):
        return None
    i_minus_where_obs[index_negative] = _np.inf
    index = _np.argmin(i_minus_where_obs)
    return where_obs[index]
    #]


def _nearest_index(i: int, where_obs: _np.ndarray, /, ) -> int:
    """
    """
    #[
    i_minus_where_obs = _np.abs(i - where_obs)
    index = _np.argmin(i_minus_where_obs)
    return where_obs[index]
    #]


_METHOD_FACTORY = {
    "next": _ft.partial(_fill_neighbor, func=_next_index, ),
    "previous": _ft.partial(_fill_neighbor, func=_previous_index, ),
    "nearest": _ft.partial(_fill_neighbor, func=_nearest_index, ),
    "constant": _fill_constant,
}





"""
"""


#[
from __future__ import annotations

from typing import (Any, Iterable, )
from numbers import (Real, )
import functools as _ft
import dataclasses as _dc
import numpy as _np
import scipy as _sp

from ..dataslates.main import (Dataslate, )
from ..databoxes.main import (Databox, )
from .solutions import (right_div, left_div, )

from ..dates import (Dater, )
from .. import quantities as _quantities
from . import initializers as _initializers
#]


class KalmanOutputData:
    """
......................................................................

==Output data from Kalman filter==


......................................................................
    """
    #[

    __slots__ = (
        "predict_mse_measurement",
        "predict_med",
        "predict_std",
        "predict_err",
        "update_med",
        "update_std",
        "smooth_med",
        "smooth_std",
    )

    def __init__(
        self: Self,
        output_store: _OutputStore,
    ) -> None:
        """
        """
        for attr_name in self.__slots__:
            attr = getattr(output_store, attr_name, )
            if hasattr(attr, "to_output_arg", ):
                attr = attr.to_output_arg()
            setattr(self, attr_name, attr, )

    def __repr__(self, /, ) -> str:
        """
        """
        slots_to_repr = (s for s in self.__slots__ if getattr(self, s) is not None)
        return f"KalmanOutputData({', '.join(slots_to_repr)})"

    #]


class _Needs:
    """
    """
    #[

    return_predict = True
    return_update = True
    return_smooth = True
    rescale_variance = False

    def __init__(self: Self, **kwargs, ) -> None:
        """
        """
        for k, v in kwargs.items():
            if not hasattr(self, k) or getattr(self, k) is None:
                continue
            setattr(self, k, v)

    #]


class _LogDataslate:
    """
    """
    #[

    @classmethod
    def from_dataslate(
        klass,
        input_ds: Dataslate,
        num_variants: int,
        name_to_log_name: dict[str, str],
        /,
    ) -> Self:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_template(input_ds, num_variants=num_variants, )
        self._prepare_log_names(name_to_log_name, )
        return self

    @classmethod
    def from_names_dates(
        klass,
        measurement_names: Iterable[str],
        dates: Iterable[Dater],
        base_columns: Iterable[int],
        num_variants: int,
        name_to_log_name: dict[str, str],
    ) -> Self:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_names_dates(
            measurement_names, dates,
            base_columns=base_columns,
            num_variants=num_variants,
        )
        self._prepare_log_names(name_to_log_name, )
        return self

    def _prepare_log_names(
        self,
        name_to_log_name: dict[str, str],
        /,
    ) -> None:
        """
        """
        self._dataslate.rename(name_to_log_name, )
        output_names = self._dataslate.output_names
        self._name_to_log_name = {
            name: log_name
            for name, log_name in name_to_log_name.items()
            if log_name in output_names
        }

    def extend(
        self,
        other: Self,
        /,
    ) -> None:
        """
        """
        self._dataslate.extend(other._dataslate, )

    #]


class _MedLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def store(
        self: Self,
        med_array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        """
        """
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            med_array = transform[rhs_indexes, :] @ med_array
        else:
            med_array = med_array[rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = med_array

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        db = self._dataslate.to_databox()
        for name, log_name in self._name_to_log_name.items():
            db[name] = db[log_name].copy()
            db[name].exp()
        return db

    #]


class _StdLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def store(
        self: Self,
        mse_array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            transform_rhs_indexes = transform[rhs_indexes, :]
            mse_array = transform_rhs_indexes @ mse_array @ transform_rhs_indexes.T
        else:
            mse_array = mse_array[rhs_indexes, ...][..., rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = _std_from_mse(mse_array, )

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        return self._dataslate.to_databox()

    #]


class _OutputStore:
    """
    """
    #[

    __slots__ = (
        "predict_med",
        "predict_std",
        "predict_mse_measurement",

        "update_med",
        "update_std",
        "predict_err",

        "smooth_med",
        "smooth_std",
    )

    def __init__(
        self,
        input_ds: Dataslate,
        num_variants: int,
        measurement_names: Iterable[str],
        needs: _Needs,
        name_to_log_name: dict[str, str] | None = None,
    ) -> Self:
        """
        """
        _med_constructor = _ft.partial(
            _MedLogDataslate.from_dataslate,
            input_ds, num_variants, name_to_log_name,
        )
        _std_constructor = _ft.partial(
            _StdLogDataslate.from_dataslate,
            input_ds, num_variants, name_to_log_name,
        )
        if needs.return_predict:
            self.predict_med = _med_constructor()
            self.predict_std = _std_constructor()
            self.predict_mse_measurement = [ () for _ in range(num_variants, ) ]
        if needs.return_update:
            self.update_med = _med_constructor()
            self.update_std = _std_constructor()
            self.predict_err = _MedLogDataslate.from_names_dates(
                measurement_names,
                input_ds.dates,
                input_ds.base_columns,
                num_variants,
                name_to_log_name,
            )
        if needs.return_smooth:
            self.smooth_med = _med_constructor()
            self.smooth_std = _std_constructor()

    def extend(self, other: Self, /, ) -> None:
        """
        """
        for s in self.__slots__:
            if getattr(self, s) is not None:
                getattr(self, s).extend(getattr(other, s), )

    def rescale_stds(self, var_scale, /, ) -> None:
        """
        """
        if var_scale == 1 or var_scale is None:
            return
        std_scale = _sqrt_positive(self.var_scale)
        for attr_name in ("predict_std", "update_std", "smooth_std", ):
            attr = getattr(self, attr_name, )
            if attr is None:
                continue
            attr._variants = [ v.data * std_scale for v in attr._variants ]

    def create_output_data(self: Self, /, ) -> KalmanOutputData:
        """
        """
        return KalmanOutputData(self, )

    #]


@_dc.dataclass(slots=True, )
class _Cache:
    """
    """
    #[
    LOG_2_PI = _np.log(2 * _np.pi)

    needs: _Needs | None = None

    base_columns: Any | None = None
    nonbase_columns: Any | None = None
    needs: Any | None = None

    all_num_obs: Any | None = None
    all_pex_invFx_pex: Any | None = None
    all_invFx_pex: Any | None = None
    all_det_Fx: Any | None = None

    all_L: Any | None = None
    all_a0: Any | None = None
    all_P0: Any | None = None
    all_ZaxT_invFx_Zax: Any | None = None
    all_ZaxT_invFx: Any | None = None
    all_G: Any | None = None
    all_inx_y: Any | None = None
    all_pex: Any | None = None
    all_invFx: Any | None = None

    all_a1: Any | None = None
    all_P1: Any | None = None

    sum_num_obs: Any | None = None
    sum_log_det_Fx: Any | None = None
    sum_pex_invFx_pex: Any | None = None

    y1_array: Any | None = None
    u0_array: Any | None = None
    v0_array: Any | None = None
    w0_array: Any | None = None

    std_u_array: Any | None = None
    std_w_array: Any | None = None
    all_cov_u: Any | None = None
    all_cov_w: Any | None = None

    var_scale: Any | None = None
    neg_log_likelihood: Any | None = None
    diffuse_factor: Any | None = None

    curr_qids: list[int] | None = None
    curr_pos: list[int] | None = None
    y_qids: list[int] | None = None
    u_qids: list[int] | None = None
    v_qids: list[int] | None = None
    w_qids: list[int] | None = None
    std_u_qids: list[int] | None = None
    std_w_qids: list[int] | None = None

    logly_within_y: list[int] | None = None

    def __init__(
        self: Self,
        input_ds: Dataslate,
        needs: _Needs,
    ) -> None:
        """
        """
        num_ext_periods = input_ds.num_periods
        self.base_columns = input_ds.base_columns
        self.nonbase_columns = input_ds.nonbase_columns
        self.needs = needs
        #
        self.all_num_obs = [None] * num_ext_periods
        self.all_pex_invFx_pex = [None] * num_ext_periods
        self.all_invFx_pex = [None] * num_ext_periods
        self.all_det_Fx = [None] * num_ext_periods
        self.all_L = [None] * num_ext_periods
        self.all_a0 = [None] * num_ext_periods
        self.all_P0 = [None] * num_ext_periods
        self.all_pex = [None] * num_ext_periods
        self.all_invFx = [None] * num_ext_periods
        self.all_ZaxT_invFx = [None] * num_ext_periods
        self.all_ZaxT_invFx_Zax = [None] * num_ext_periods
        self.all_G = [None] * num_ext_periods
        self.all_a1 = [None] * num_ext_periods
        self.all_P1 = [None] * num_ext_periods
        self.all_cov_u = [None] * num_ext_periods
        self.all_cov_w = [None] * num_ext_periods
        #
        self.var_scale = None
        self.neg_log_likelihood = None
        self.diffuse_factor = None
        #
        self.curr_qids = None
        self.curr_pos = None
        self.y_qids = None
        self.u_qids = None
        self.v_qids = None
        self.w_qids = None
        self.std_u_qids = None
        self.std_w_qids = None

    @property
    def reversed_base_columns(self: Self, /, ) -> Any:
        """
        """
        return reversed(self.base_columns, )

    def calculate_likelihood(self: Self, /, ) -> None:
        """
        """
        self.sum_num_obs = sum(i for i in self.all_num_obs if i is not None)
        self.sum_log_det_Fx = sum(i for i in self.all_det_Fx if i is not None)
        self.sum_pex_invFx_pex = sum(i for i in self.all_pex_invFx_pex if i is not None)
        #
        self._calculate_variance_scale()
        #
        self.neg_log_likelihood = (
            + self.sum_num_obs*self.LOG_2_PI
            + self.sum_log_det_Fx
            + self.sum_pex_invFx_pex
        ) / 2;

    def _calculate_variance_scale(self: Self, /, ) -> None:
        """
        """
        if not self.needs.rescale_variance:
            self.var_scale = 1
            return
        if self.sum_num_obs == 0:
            self.var_scale = 1
            self.sum_pex_invFx_pex = 0
            return
        self.var_scale = self.sum_pex_invFx_pex / self.sum_num_obs
        self.sum_log_det_Fx += self.sum_num_obs * _np.log(self.var_scale)
        self.sum_pex_invFx_pex = self.sum_pex_invFx_pex / self.var_scale

    def create_output_info(self: Self, /, ) -> dict[str, Any]:
        """
        """
        return {
            "neg_log_likelihood": self.neg_log_likelihood,
            "var_scale": self.var_scale,
            "diffuse_factor": self.diffuse_factor,
            "cache": self,
        }

    #]


def _sqrt_positive(x):
    return _np.sqrt(_np.maximum(x, 0.0, ))


def _std_from_mse(mse, ):
    return _sqrt_positive(_np.diag(mse).copy())


class Mixin:
    """
    """
    #[

    def kalman_filter(
        self,
        #
        input_db: Databox,
        span: Iterable[Dater],
        #
        diffuse_factor: Real | None = None,
        #
        return_: Iterable[str, ...] = ("predict", "update", "smooth", ),
        return_predict: bool = True,
        return_update: bool = True,
        return_smooth: bool = True,
        rescale_variance: bool = False,
        #
        shocks_from_databox: bool = False,
        stds_from_databox: bool = False,
    ) -> Databox:
        """
        """
        #[

        work_db = input_db.shallow()
        if not shocks_from_databox:
            shock_names = self.get_names(kind=_quantities.ANY_SHOCK, )
            work_db.remove(shock_names, strict_names=False, )
        if not stds_from_databox:
            std_names = self.get_names(kind=_quantities.ANY_STD, )
            work_db.remove(std_names, strict_names=False, )

        input_ds = Dataslate.from_databox_for_slatable(
            self,
            work_db,
            span,
            extend_span=False,
        )

        needs = _Needs(
            return_predict=return_predict and "predict" in return_,
            return_update=return_update and "update" in return_,
            return_smooth=return_smooth and "smooth" in return_,
            rescale_variance=rescale_variance,
        )

        solution_vectors = self._solution_vectors
        temp = [(t.qid, i) for i, t in enumerate(solution_vectors.transition_variables) if not t.shift ]
        curr_qids, curr_pos = tuple(zip(*temp))
        curr_qids = list(curr_qids)
        curr_pos = list(curr_pos)

        y_qids = [ t.qid for t in solution_vectors.measurement_variables ]
        u_qids = [ t.qid for t in solution_vectors.unanticipated_shocks ]
        v_qids = [ t.qid for t in solution_vectors.anticipated_shocks ]
        w_qids = [ t.qid for t in solution_vectors.measurement_shocks ]

        qid_to_logly = self.create_qid_to_logly()
        logly_within_y = tuple( i for i, qid in enumerate(y_qids, ) if qid_to_logly.get(qid, False) )

        std_u_qids = self.get_std_qids_for_shock_qids(u_qids, )
        std_w_qids = self.get_std_qids_for_shock_qids(w_qids, )

        num_variants = 1

        qid_to_name = self.create_qid_to_name()
        y_names = [ qid_to_name[qid] for qid in y_qids ]

        name_to_log_name = {
            n: _quantities.wrap_logly(n, )
            for qid, n in enumerate(input_ds.names, )
            if qid_to_logly.get(qid, False, )
        } if qid_to_logly else {}

        output_store = _OutputStore(
            input_ds=input_ds,
            num_variants=0,
            measurement_names=y_names,
            needs=needs,
            name_to_log_name=name_to_log_name,
        )

        output_info = []


        #
        # Variant dependent from here
        #
        input_ds_v = input_ds.get_variant(0, )
        solution = self.get_solution_matrices()

        output_store_v = _OutputStore(
            input_ds=input_ds,
            num_variants=1,
            measurement_names=y_names,
            needs=needs,
            name_to_log_name=name_to_log_name,
        )

        cache = _Cache(
            input_ds=input_ds,
            needs=needs,
        )

        cache.curr_qids = list(curr_qids)
        cache.curr_pos = list(curr_pos)
        cache.y_qids = list(y_qids)
        cache.u_qids = list(u_qids)
        cache.v_qids = list(v_qids)
        cache.w_qids = list(w_qids)
        cache.std_u_qids = list(std_u_qids)
        cache.std_w_qids = list(std_w_qids)
        cache.logly_within_y = logly_within_y

        #
        # Calculate initial med and mse
        #
        init_cov_u = self.get_cov_unanticipated_shocks()
        init_med, init_mse, cache.diffuse_factor = _initializers.initialize(
            solution,
            init_cov_u,
            diffuse_scale=diffuse_factor,
        )

        # tolerance = 1e-12
        # Za[_np.abs(Za)<tolerance] = 0
        # Ua[_np.abs(Ua)<tolerance] = 0
        # Pa[_np.abs(Pa)<tolerance] = 0

        input_data_array = input_ds_v.get_data_variant()


        a1, P1 = init_med, init_mse
        Ta = solution.Ta
        Ra = solution.Ra
        Pa = solution.Pa
        Ka = solution.Ka
        Za = solution.Za
        H = solution.H
        D = solution.D
        Ua = solution.Ua

        cache.y1_array = input_data_array[cache.y_qids, :]
        cache.u0_array = input_data_array[cache.u_qids, :]
        cache.v0_array = input_data_array[cache.v_qids, :]
        cache.w0_array = input_data_array[cache.w_qids, :]
        cache.std_u_array = input_data_array[cache.std_u_qids, :]
        cache.std_w_array = input_data_array[cache.std_w_qids, :]

        if cache.logly_within_y:
            cache.y1_array[cache.logly_within_y, :] = _np.log(cache.y1_array[cache.logly_within_y, :])

        for n in ("y1_array", "u0_array", "v0_array", "w0_array", ):
            array = getattr(cache, n, )
            array[:, cache.nonbase_columns] = _np.nan
            setattr(cache, n, array, )

        cache.all_inx_y = tuple(
            _np.isfinite(column_data, )
            for column_data in cache.y1_array.T
        )

        if needs.return_predict:
            output_store_v.predict_med.store(cache.u0_array, (cache.u_qids, ...), )
            output_store_v.predict_med.store(cache.v0_array, (cache.v_qids, ...), )
            output_store_v.predict_med.store(cache.w0_array, (cache.w_qids, ...), )

        if needs.return_update:
            output_store_v.update_med.store(cache.y1_array, (cache.y_qids, ...), )

        if needs.return_smooth:
            pass

        for t in cache.base_columns:

            inx_y = cache.all_inx_y[t]

            # Extract time-varying means and stds for shocks
            u0 = cache.u0_array[:, t]
            v0 = cache.v0_array[:, t]
            w0 = cache.w0_array[:, t]
            cov_u = _np.diag(cache.std_u_array[:, t]**2, )
            cov_w = _np.diag(cache.std_w_array[:, t]**2, )

            # MSE prediction step
            P0 = Ta @ P1 @ Ta.T + Pa @ cov_u @ Pa.T
            P0 = (P0 + P0.T) / 2
            F = Za @ P0 @ Za.T + H @ cov_w @ H.T
            F = (F + F.T) / 2
            Fx = F[inx_y, :][:, inx_y]
            if needs.return_predict:
                output_store_v.predict_std.store(P0, (cache.curr_qids, t), rhs_indexes=cache.curr_pos, transform=Ua, )
                output_store_v.predict_mse_measurement[0] += (Fx, )

            # Mean prediction step
            a0 = Ta @ a1 + Ka + Pa @ u0
            y0 = Za @ a0 + D + H @ w0
            y0[~inx_y] = _np.nan

            if needs.return_predict:
                output_store_v.predict_med.store(a0, (cache.curr_qids, t), rhs_indexes=cache.curr_pos, transform=Ua, )
                output_store_v.predict_med.store(y0, (cache.y_qids, t), )
                output_store_v.predict_med.store(u0, (cache.u_qids, t), )
                output_store_v.predict_med.store(v0, (cache.v_qids, t), )
                output_store_v.predict_med.store(w0, (cache.w_qids, t), )

            # MSE updating step
            Zax = Za[inx_y, :]
            # ZaxT_invFx = right_div(Zax.T, Fx, )
            # if Fx.size > 0:
            #     sing_values = _np.linalg.svd(Fx, compute_uv=False, )
            #     relative_cond = sing_values[-1] / sing_values[0]
            #     print(t, Fx.size, relative_cond, sing_values[0], )
            invFx = _np.linalg.inv(Fx)
            ZaxT_invFx = Zax.T @ invFx
            G = P0 @ ZaxT_invFx
            P1 = P0 - G @ Zax @ P0
            P1 = (P1 + P1.T) / 2

            if needs.return_update:
                output_store_v.update_std.store(P1, (cache.curr_qids, t), rhs_indexes=cache.curr_pos, transform=Ua, )

            # Mean updating step
            cache.all_num_obs[t] = _np.count_nonzero(inx_y, )
            y1 = cache.y1_array[:, t]
            pe = y1 - y0
            pex = pe[inx_y]
            a1 = a0 + G @ pex

            # invFx_pex = left_div(Fx, pex, ) # inv(Fx) * pex
            invFx_pex = invFx @ pex # inv(Fx) * pex

            if needs.return_update:
                Hx_cov_w = H[inx_y, :] @ cov_w
                w1 = w0 + Hx_cov_w.T @ invFx_pex
                Pa_cov_u = Pa @ cov_u
                r = ZaxT_invFx @ pex # Z' * inv(F) * pe
                u1 = u0 + Pa_cov_u.T @ r
                v1 = v0
                output_store_v.update_med.store(a1, (cache.curr_qids, t), rhs_indexes=cache.curr_pos, transform=Ua, )
                output_store_v.update_med.store(y1, (cache.y_qids, t), )
                output_store_v.update_med.store(u1, (cache.u_qids, t), )
                output_store_v.update_med.store(v1, (cache.v_qids, t), )
                output_store_v.update_med.store(w1, (cache.w_qids, t), )
                output_store_v.predict_err.store(pe, (..., t), )

            if needs.return_smooth:
                cache.all_a0[t] = a0
                cache.all_P0[t] = P0
                cache.all_a1[t] = a1
                cache.all_P1[t] = P1
                cache.all_invFx_pex[t] = invFx_pex
                cache.all_ZaxT_invFx[t] = Zax.T @ invFx
                cache.all_ZaxT_invFx_Zax[t] = ZaxT_invFx @ Zax
                cache.all_L[t] = Ta - (Ta @ G) @ Zax
                cache.all_G[t] = G
                cache.all_cov_u[t] = cov_u
                cache.all_cov_w[t] = cov_w

            cache.all_pex[t] = pex
            cache.all_invFx[t] = invFx
            cache.all_pex_invFx_pex[t] = pex @ invFx_pex
            cache.all_det_Fx[t] = _np.log(_np.linalg.det(Fx))


        if needs.return_smooth:
            _smooth_backward(cache, output_store_v, needs, solution, )

        cache.calculate_likelihood()
        output_store_v.rescale_stds(cache.var_scale, )

        output_store.extend(output_store_v, )
        output_info.append(cache.create_output_info(), )

        return output_store.create_output_data(), output_info

    #]


def _smooth_backward(
    cache: _Cache,
    output_store_v: _OutputStore,
    needs: _Needs,
    solution: _simultaneous.Solution,
) -> None:
    """
    """
    #[

    Ta = solution.Ta
    Ua = solution.Ua
    Pa = solution.Pa
    H = solution.H

    N = None
    r = None

    for t in cache.reversed_base_columns:

        inx_y = cache.all_inx_y[t]
        u0 = cache.u0_array[:, t]
        v0 = cache.v0_array[:, t]
        w0 = cache.w0_array[:, t]
        cov_u = cache.all_cov_u[t]
        cov_w = cache.all_cov_w[t]

        a0 = cache.all_a0[t]
        L = cache.all_L[t]
        G = cache.all_G[t]
        P0 = cache.all_P0[t]
        ZaxT_invFx = cache.all_ZaxT_invFx[t]
        ZaxT_invFx_Zax = cache.all_ZaxT_invFx_Zax[t]
        invFx = cache.all_invFx[t]
        pex = cache.all_pex[t]

        N = ZaxT_invFx_Zax + ((L.T @ N @ L) if N is not None else 0)
        P2 = P0 - P0 @ N @ P0
        output_store_v.smooth_std.store(P2, (cache.curr_qids, t), rhs_indexes=cache.curr_pos, transform=Ua, )

        H_cov_w = H[inx_y, :] @ cov_w
        w2 = w0 + H_cov_w.T @ (invFx @ pex - (((Ta @ G).T @ r) if r is not None else 0))
        output_store_v.smooth_med.store(w2, (cache.w_qids, t), )

        r = ZaxT_invFx @ pex + ((L.T @ r) if r is not None else 0)
        a2 = a0 + P0 @ r
        output_store_v.smooth_med.store(a2, (cache.curr_qids, t), rhs_indexes=cache.curr_pos, transform=Ua, )

        y2 = cache.y1_array[:, t]
        output_store_v.smooth_med.store(y2, (cache.y_qids, t), )

        Pa_cov_u = Pa @ cov_u
        u2 = u0 + Pa_cov_u.T @ r
        output_store_v.smooth_med.store(u2, (cache.u_qids, t), )

        v2 = v0
        output_store_v.smooth_med.store(v2, (cache.v_qids, t), )

    #]




"""
Differentiation and invariation rules
"""

#[
from __future__ import annotations
#]

def add_diff(self, other):
    """
    Differentiate self(x)+other(x) or self(x)+other
    """
    if hasattr(other, "_is_atom"):
        new_value = self.value + other.value
        new_diff = self.diff + other.diff
    else:
        new_value = self.value + other
        new_diff = self.diff
    return type(self).no_context(new_value, new_diff, False)

def add_invar(self, other):
    """
    Invariance of self(x)+other(x) or self(x)+other
    """
    if hasattr(other, "_is_atom"):
        new_diff = self._diff | other._diff
        new_invariant = self._invariant & other._invariant
    else:
        new_diff = self._diff
        new_invariant = self._invariant
    return Atom.no_context(new_diff, new_invariant, False)



def sub_diff(self, other):
    if hasattr(other, "_is_atom"):
        new_value = self.value - other.value
        new_diff = self.diff - other.diff
    else:
        new_value = self.value - other
        new_diff = self.diff
    new_logly = False
    return type(self).no_context(new_value, new_diff, False)

sub_invar = add_invar





"""
Temporal change functions
"""


#[
from __future__ import annotations

from typing import (Callable, )
from numbers import (Real, )
import numpy as _np

from .. import dates as _dates
from .. import pages as _pages
from . import main as _series
from . import _functionalize
#]


__all__ = ()


class Inlay:
    """
    """
    #[

    @_pages.reference(
        category=None,
        call_name="Temporal change calculations",
        call_name_is_code=False,
        priority=20,
    )
    def temporal_change(
        self,
        shift: int | str,
        func: Callable,
        /,
    ) -> None:
        r"""
················································································


Overview of temporal change functions:

| Function      | Description
|---------------|-------------
| `diff`        | First difference, $y_t = x_t - x_{t-k}$ 
| `diff_log`    | First difference of logs, $y_t = \log x_t - \log x_{t-k}$ 
| `pct`         | Percent change, $y_t = 100 (x_t/x_{t-k} - 1)$ 
| `roc`         | Gross rate of change, $y_t = x_t/x_{t-k}$ 
| `adiff`       | Annualized first difference, $y_t = a \cdot (x_t - x_{t-1})$ 
| `adiff_log`   | Annualized first difference of logs, $y_t = a \cdot (\log x_t - \log x_{t-1})$ 
| `apct`        | Annualized percent change, $y_t = 100 [(x_t/x_{t-1})^a - 1]$ 
| `aroc`        | Annualized gross rate of change, $y_t = (x_t/x_{t-1})^a$ 

where

* $k$ is the time shift (time lag) with which the temporal change is
calculated, determined by the negative value of the `shift` input argument;

* $a$ is the annualization factor, determined by the frequency of the time
series (i.e. the number of calendar periods within a year).


### Functional forms creating a new Series object ###


```
new = irispie.diff(self, /, shift=-1)
new = irispie.diff_log(self, /, shift=-1)
new = irispie.pct(self, /, shift=-1)
new = irispie.roc(self, /, shift=-1)
new = irispie.adiff(self)
new = irispie.adiff_log(self)
new = irispie.apct(self)
new = irispie.aroc(self)
```


### Class methods changing an existing Series object in-place ###


```
self.diff(shift=-1)
self.difflog(shift=-1)
self.pct(shift=-1)
self.roc(shift=-1)
self.adiff()
self.adifflog()
self.apct()
self.aroc()
```



### Input arguments ###


???+ input "self"
    Time series on whose data a temporal change function is calculated (see
    the overview table above).

???+ input "shift"
    A negative integer determining a time lag at which the temporal change
    function is calculated. If `shift=None` (or not specified), `shift` is
    set to `-1`.


### Returns ###


???+ returns "new"
    New time series with data calculated as a temporal change function of
    the original data.

???+ returns "self"
    Time series with data replaced by a temporal change function of the
    original data.

················································································
        """
        _catch_invalid_shift(shift, )
        other = self.copy()
        other.shift(shift, )
        self._binop(other, func, new=self, )

    @_pages.reference(category="temporal_change", )
    def diff(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
················································································

==First difference==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        self.temporal_change(shift, lambda x, y: x - y, )

    @_pages.reference(category="temporal_change", )
    def adiff(
        self,
        /,
    ) -> None:
        r"""
················································································

==Annualized first difference==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: factor*(x - y), )

    @_pages.reference(category="temporal_change", )
    def diff_log(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
················································································

==First difference of logs==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        self.temporal_change(shift, lambda x, y: _np.log(x) - _np.log(y), )

    @_pages.reference(category="temporal_change", )
    def adiff_log(
        self,
        /,
    ) -> None:
        r"""
················································································

==Annualized first difference of logs==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: factor*(_np.log(x) - _np.log(y)), )

    @_pages.reference(category="temporal_change", )
    def roc(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
················································································

==Gross rate of change==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        self.temporal_change(shift, lambda x, y: x/y, )

    @_pages.reference(category="temporal_change", )
    def aroc(
        self,
        /,
    ) -> None:
        r"""
················································································

==Annualized gross rate of change==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: (x/y)**factor, )

    @_pages.reference(category="temporal_change", )
    def pct(
        self,
        shift: int | str = -1,
        /,
    ) -> None:
        r"""
················································································

==Percent change==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        self.temporal_change(shift, lambda x, y: 100*(x/y - 1), )

    @_pages.reference(category="temporal_change", )
    def apct(
        self,
        /,
    ) -> None:
        r"""
················································································

==Annualized percent change==

See documentation of [temporal change calculations](#temporal-change-calculations) or
`help(irispie.Series.temporal_change)`.

················································································
        """
        shift = -1
        factor = self.frequency.value or 1
        self.temporal_change(shift, lambda x, y: 100*((x/y)**factor - 1), )

    def roc_from_pct(
        self,
        /,
    ) -> None:
        r"""
        """
        self.data = 1 + self.data/100

    def pct_from_roc(
        self,
        /,
    ) -> None:
        r"""
        """
        self.data = 100*(self.data - 1)

    def pct_from_apct(
        self,
        /,
    ) -> None:
        r"""
        """
        factor = self.frequency.value or 1
        self.data = 100*((1 + self.data/100)**(1/factor) - 1)

    def roc_from_apct(
        self,
        /,
    ) -> None:
        r"""
        """
        factor = self.frequency.value or 1
        self.data = (1 + self.data/100)**(1/factor)

    def roc_from_aroc(
        self,
        /,
    ) -> None:
        r"""
        """
        factor = self.frequency.value or 1
        self.data = self.data**factor

    @_pages.reference(category="temporal_cumulation", )
    def cum_diff(self, *args, **kwargs, ) -> None:
        r"""
················································································

==Cumulation of first differences==

See documentation of [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

················································································
        """
        self.temporal_cumulation("diff", *args, **kwargs, )

    @_pages.reference(category="temporal_cumulation", )
    def cum_diff_log(self, *args, **kwargs, ) -> None:
        r"""
················································································

==Cumulation of first differences of logs==

See documentation of [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

················································································
        """
        self.temporal_cumulation("diff_log", *args, **kwargs, )

    @_pages.reference(category="temporal_cumulation", )
    def cum_pct(self, *args, **kwargs, ) -> None:
        r"""
················································································

==Cumulation of percent changes==

See documentation of [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

················································································
        """
        self.temporal_cumulation("pct", *args, **kwargs, )

    @_pages.reference(category="temporal_cumulation", )
    def cum_roc(self, *args, **kwargs, ) -> None:
        r"""
················································································

==Cumulation of gross rates of change==

See documentation of [temporal cumulation calculations](#temporal-cumulation-calculations) or
`help(irispie.Series.temporal_cumulation)`.

················································································
        """
        self.temporal_cumulation("roc", *args, **kwargs, )

    @_pages.reference(
        category=None,
        call_name="Temporal cumulation calculations",
        call_name_is_code=False,
        priority=20,
    )
    def temporal_cumulation(
        self,
        func_name: str,
        shift: int | str = -1,
        initial: Real | _series.Series | None = None,
        span: _dates.Ranger | None = None,
    ) -> None:
        r"""
················································································

Overview of temporal cumulation calculations:

| Function      | Description
|---------------|-------------
| `cum_diff`    | Cumulative difference, $y_t = \sum_{i=0}^{t} (x_i - x_{i-k})$
| `cum_diff_log`| Cumulative difference of logs, $y_t = \sum_{i=0}^{t} (\log x_i - \log x_{i-k})$
| `cum_pct`     | Cumulative percent change, $y_t = 100 \sum_{i=0}^{t} (x_i/x_{i-k} - 1)$
| `cum_roc`     | Cumulative gross rate of change, $y_t = \prod_{i=0}^{t} (x_i/x_{i-k})$

where

* $k$ is the time shift (time lag) with which the temporal change is
calculated, determined by the negative value of the `shift` input argument.

* $t$ is the end date of the time series.


### Functional forms creating new Series objects ###


```
new = irispie.cum_diff(self, /, shift=-1, initial=None, span=None)
new = irispie.cum_diff_log(self, /, shift=-1, initial=None, span=None)
new = irispie.cum_pct(self, /, shift=-1, initial=None, span=None)
new = irispie.cum_roc(self, /, shift=-1, initial=None, span=None)
```


### Class methods changing existing Series objects in-place ###


```
self.cum_diff(/, shift=-1, initial=None, span=None)
self.cum_diff_log(/, shift=-1, initial=None, span=None)
self.cum_pct(/, shift=-1, initial=None, span=None)
self.cum_roc(/, shift=-1, initial=None, span=None)
```


### Input arguments ###


???+ input "self"
    Time series on whose data a temporal cumulation is calculated (see the
    overview table above).

???+ input "shift"
    A negative integer determining a time lag at which the temporal
    cumulation is calculated. If `shift=None` (or not specified), `shift`
    is set to `-1`.

???+ input "initial"
    Initial value of the cumulative series. If `initial=None` (or not specified),
    the initial value is set to `0` for `diff` and `diff_log`, and to
    `1` for `pct` and `roc`.

???+ input "span"
    Time span on which the values from the original series are cumulated.
    If `span=None` (or not specified), the time span is set to the entire
    time series.


### Returns ###


???+ returns "new"
    New time series with data calculated as temporal cumulation of the
    original data.

???+ returns "self"
    Time series with data replaced by temporal cumulation of the original
    data.


················································································
        """
        _catch_invalid_shift(shift, )
        span = _dates.Ranger(None, None, ) if span is None else span
        span = span.resolve(self, )
        direction = span.direction
        factory = _CUMULATIVE_FACTORY[func_name]
        cum_func = factory[direction]
        initial = factory["initial"] if initial is None else initial
        if direction == "forward":
            self._cumulate_forward(shift, cum_func, initial, span, )
        elif direction == "backward":
            self._cumulate_backward(shift, cum_func, initial, span, )

    def _cumulate_forward(self, shift, cum_func, initial, span, /, ) -> None:
        """
        """
        shifted_range = tuple(t.shift(shift, ) for t in span)
        initial_range = _dates.Ranger(min(shifted_range), span.end_date, )
        orig = self.copy()
        self.empty()
        self.set_data(initial_range, initial)
        for t, sh in zip(span, shifted_range):
            new_data = cum_func(self.get_data(sh, ), orig.get_data(t, ), )
            self.set_data(t, new_data)

    def _cumulate_backward(self, shift, cum_func, initial, shifted_backward_range, /, ) -> None:
        """
        """
        orig_range_shifted = _dates.Ranger(self.start_date, self.end_date, -1, )
        orig_range_shifted.shift(shift, )
        shifted_backward_range = shifted_backward_range.resolve(orig_range_shifted, )
        backward_range = shifted_backward_range.copy()
        backward_range.shift(-shift, )
        initial_range = _dates.Ranger(min(shifted_backward_range), backward_range.start_date, )
        orig = self.copy()
        self.empty()
        self.set_data(initial_range, initial, )
        for t, sh in zip(backward_range, shifted_backward_range, ):
            new_data = cum_func(self.get_data(t, ), orig.get_data(t, ), )
            self.set_data(sh, new_data, )

    #]


attributes = (n for n in dir(Inlay) if not n.startswith("_"))
for n in attributes:
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


_CUMULATIVE_FACTORY = {
    "diff": {
        "forward": lambda x_past, change_curr: x_past + change_curr,
        "backward": lambda x_future, change_future: x_future - change_future,
        "initial": 0,
    },
    "diff_log": {
        "forward": lambda x_past, change_curr: x_past * exp(change_curr),
        "backward": lambda x_future, change_future: x_future / exp(change_future),
        "initial": 0,
    },
    "pct": {
        "forward": lambda x_past, change_curr: x_past * (1 + change_curr/100),
        "backward": lambda x_future, change_future: x_future / (1 + change_future/100),
        "initial": 1,
    },
    "roc": {
        "forward": lambda x_past, change_curr: x_past * change_curr,
        "backward": lambda x_future, change_future: x_future / change_future,
        "initial": 1,
    },
}


def _catch_invalid_shift(shift: int | str, ):
    """
    """
    if not isinstance(shift, str) and (int(shift) != shift or shift >= 0):
        raise ValueError("Time shift must be a negative integer or a string")


def _roc_from_pct(pct: Real, /, ) -> Real:
    """
    """
    return 1 + pct/100





"""
Initialize mean and MSE matrix for alpha vector
"""


#[
from __future__ import annotations

import numpy as _np
from numbers import (Real, )

from . import solutions as _solutions
from . import covariances as _covariances
#]


_DEFAULT_DIFFUSE_SCALE = 1e7


def initialize(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
    mse_method: Literal["fixed", "approx_diffuse", ] = "approx_diffuse",
    diffuse_scale: Real | None = None,
) -> tuple[np_.ndarray, np_.ndarray]:
    """
    """
    diffuse_scale = (
        diffuse_scale if diffuse_scale is not None
        else _DEFAULT_DIFFUSE_SCALE
    )
    if mse_method == "fixed":
        diffuse_scale = 0.0
    return (
        _initialize_mean(solution, ),
        _initialize_mse(solution, cov_u, diffuse_scale, ),
        diffuse_scale,
    )


def _initialize_mean(
    solution: _solutions.Solution,
    /,
) -> _np.ndarray:
    """
    Solve alpha_stable = Ta_stable @ alpha_stable + Ka_stable for alpha_stable
    and return alpha with 0s for unstable elements
    """
    #[
    num_alpha = solution.num_alpha
    num_unit_roots = solution.num_unit_roots
    num_stable = solution.num_stable
    Ta_stable = solution.Ta_stable
    Ka_stable = solution.Ka_stable
    init_mean = _np.zeros((num_alpha, ), dtype=float, )
    #
    T = _np.eye(num_stable, dtype=float, ) - Ta_stable
    init_mean[num_unit_roots:] = _solutions.left_div(T, Ka_stable, )
    return init_mean
    #]


def _initialize_mse(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
    diffuse_scale: Real | None = None,
    /,
) -> _np.ndarray:
    """
    """
    #[
    init_mse = _covariances.get_cov_alpha_00(solution, cov_u, )
    if diffuse_scale:
        num_unit_roots = solution.num_unit_roots
        init_mse[:num_unit_roots, :num_unit_roots] = \
            _initialize_mse_unstable_approx_diffuse(solution, cov_u, init_mse, diffuse_scale, )
    return init_mse
    #]


def _initialize_mse_unstable_approx_diffuse(
    solution: _solutions.Solution,
    cov_u: _np.ndarray,
    init_mse: _np.ndarray,
    diffuse_scale: Real | None = None,
    /,
) -> _np.ndarray:
    """
    """
    #[
    num_unit_roots = solution.num_unit_roots
    num_alpha = solution.num_alpha
    base_cov = (
        init_mse[num_unit_roots:, num_unit_roots:] if num_alpha
        else (cov_u if cov_u.size else _np.ones((1, 1, ), dtype=float, ))
    )
    scale = _mean_of_diag(base_cov, ) * diffuse_scale
    return scale * _np.eye(num_unit_roots, dtype=float, )
    #]


def _mean_of_diag(
    x: _np.ndarray,
    /,
) -> Real:
    """
    """
    return _np.mean(_np.diag(x, ), )




"""
Import time series data from a CSV file
"""


#[
from __future__ import annotations

from typing import (Self, TypeAlias, Protocol, Callable, )
from collections.abc import (Iterator, Iterable, Generator, )
import csv as _cs
import numpy as _np
import itertools as _it
import functools as _ft

from ..series import main as _series
from .. import dates as _dates

from . import main as _databoxes
#]


_Header: TypeAlias = tuple[str, str]
_DataVector: TypeAlias = tuple[str, ...]
_HeaderIterator: TypeAlias = Iterator[_Header, ]
_DataIterator: TypeAlias = Iterator[_DataVector, ]
_DataArrayReader: TypeAlias = Callable[[Iterable[int], ], _np.ndarray]


_GENFROMTXT_SETTINGS = {
    "delimiter": ",",
    "ndmin": 2,
}


class _SheetFileFactory(Protocol):
    """
    Protocol for file factory
    """
    #[

    def create_header_and_data_iterators(self, ) -> tuple[_HeaderIterator, _DataIterator, ]:
        ...
    def create_data_array_reader(self, ) -> _DataArrayReader:
        ...

    #]


#
# FIXME: Make read_csv a class method
#
def read_csv(
    file_path: str,
    /,
    **kwargs,
) -> Iterable[Self]:
    """
    Read a CSV file
    """
    #[

    factory = _ColumnwiseFileFactory(file_path, **kwargs, )
    #
    header_iterator, data_iterator = factory.create_header_and_data_iterators()
    data_array_reader = factory.create_data_array_reader()
    all_databoxes = tuple(
        b.create_databox_from_block(data_array_reader, **kwargs, )
        for b in _generate_blocks(header_iterator, data_iterator, **kwargs, )
    )
    return all_databoxes

    #]


def _remove_nonascii_from_start(
    string: str,
    /,
) -> str:
    """
    Remove non-ASCII characters from start of string
    """
    #[
    while string and not string[0].isascii():
        string = string[1:]
    return string
    #]


def _generate_blocks(
    header_iterator: _HeaderIterator,
    data_iterator: _DataIterator,
    /,
    **kwargs,
) -> _Block | None:
    """
    Generate blocks of data of same date frequency
    """
    #[
    def _is_yieldable(block, /, ) -> bool:
        return block is not None and block.num_columns > 0
    #
    current_block = None
    for index, (header, data, ) in enumerate(zip(header_iterator, data_iterator, )):
        if _is_end_of_file(header, ):
            if _is_yieldable(current_block, ):
                yield current_block
            current_block = None
        elif _is_start_of_block(header, ):
            if _is_yieldable(current_block, ):
                yield current_block
            current_block = _Block(index, header, data, **kwargs, )
        elif current_block is not None:
            current_block.add_header(header, )
        else:
            current_block = None
    if _is_yieldable(current_block, ):
        yield current_block
    #]


class _Block:
    """
    """
    #[

    __slots__ = (
        "_start_index",
        "_headers",
        "_dates",
        "_data_array",
    )

    def __init__(
        self,
        start_index: int | None,
        header: _Header | None,
        date_str_vector: _DataVector | None,
        /,
        **kwargs,
    ) -> None:
        self._start_index = start_index
        self._headers = []
        freq = _dates.Frequency.from_letter(header[0]) if header is not None else None
        self._dates = _create_dates_for_block(freq, date_str_vector, **kwargs, )
        self._data_array = None

    def add_header(
        self,
        header: _Header,
        /,
    ) -> None:
        """
        Add a header to the block
        """
        self._headers.append(header)

    def create_databox_from_block(
        self,
        data_array_reader: _DataArrayReader,
        /,
        **kwargs,
    ) -> _databoxes.Databox:
        """
        Create a databox from the block
        """
        self._load_data_array(data_array_reader, )
        return self._data_array

    def _load_data_array(
        self,
        data_array_reader: _DataArrayReader,
        /,
    ) -> _np.ndarray:
        """
        Load block data as a numpy array
        """
        self._data_array = data_array_reader(usecols=self.column_indices, )

    @property
    def num_columns(
        self,
        /,
    ) -> int:
        """
        Return the number of columns
        """
        return len(self._headers)

    @property
    def column_indices(
        self,
        /,
    ) -> range:
        """
        Return the column indices
        """
        return range(
            self._start_index,
            self._start_index + self.num_columns,
        )

    #]


_END_OF_FILE = "__eof__"


def _is_start_of_block(
    header: _Header,
    /,
) -> bool:
    """
    Return true if the name is the start of a block
    """
    return header[0].startswith("__") and header[0] != _END_OF_FILE


def _is_end_of_file(
    header: _Header,
    /,
) -> bool:
    """
    Return true if the name is the end of a file
    """
    return header[0] == _END_OF_FILE


class _ColumnwiseFileFactory(_SheetFileFactory):
    """
    Iterator factory for columnwise data
    """
    #[
    def __init__(
        self,
        file_path: str,
        /,
        **kwargs,
    ) -> None:
        self._file_path = file_path
        self._skip_rows = kwargs.get("skip_rows", 0, )
        self._has_description_row = kwargs.get("has_description_row", False, )

    def create_header_and_data_iterators(
        self,
        /,
    ) -> tuple[Iterator[_Header], Iterator[_DataVector], ]:
        with open(self._file_path, encoding="utf-8-sig", ) as file:
            csv_reader = _cs.reader(file, )
            header_iterator = self._create_header_iterator(csv_reader, )
            data_iterator = self._create_data_iterator(csv_reader, )
        return header_iterator, data_iterator, 

    def create_data_array_reader(
        self,
        /,
    ) -> Callable[[Iterable[int], ], _np.ndarray]:
        """
        Create a reader for the data array
        """
        skip_rows = self._skip_rows + 1 + int(self._has_description_row)
        return _ft.partial(
            _np.genfromtxt,
            self._file_path,
            skip_header=skip_rows,
            **_GENFROMTXT_SETTINGS,
        )

    def _create_header_iterator(
        self,
        csv_reader,
        /,
    ) -> Iterator[_Header]:
        """
        Create an iterator over headers (name, description, )
        """
        for _ in range(self._skip_rows):
            next(csv_reader, )
        name_row = next(csv_reader, )
        if name_row and name_row[0]:
            name_row[0] = _remove_nonascii_from_start(name_row[0], )
        description_row = next(csv_reader, ) if self._has_description_row else _it.repeat("", )
        return zip(name_row, description_row, )

    def _create_data_iterator(
        self,
        csv_reader,
        /,
    ) -> Iterator[_DataVector]:
        """
        Create an iterator over individual data series or dates
        """
        return _it.zip_longest(*csv_reader, fillvalue="", )


def _create_dates_for_block(
    freq: _dates.Frequency | None,
    date_str_vector: _DataVector | None,
    /,
    start_date_only: bool = False,
    **kwargs,
) -> tuple[Dater]:
    """
    Create dates for a block
    """
    #[
    if start_date_only:
        dates_factory = _create_dates_from_start_date
    else:
        dates_factory = _create_dates_from_date_column
    return dates_factory(freq, date_str_vector, )
    #]


def _create_dates_from_start_date(
    freq: _dates.Frequency | None,
    date_str_vector: _DataVector | None,
) -> tuple[Dater]:
    """
    Create dates from a start date
    """
    #[
    num_dates = len(date_str_vector)
    start_date = _dates.Dater.from_sdmx_string(freq, date_str_vector[0], )
    return tuple(_dates.Ranger(start_date, num_dates, ))
    #]


def _create_dates_from_date_column(
    freq: _dates.Frequency | None,
    date_str_vector: _DataVector | None,
) -> tuple[Dater]:
    """
    Create dates from a date column
    """
    #[
    return tuple(
        _dates.Dater.from_sdmx_string(freq, s, ) if s else None
        for s in date_str_vector
    )
    #]




"""
Jacobians for dynamic period-by-period systems
"""


#[
from __future__ import annotations

from typing import (Protocol, )
from collections.abc import (Iterable, )
import numpy as _np

from ..incidences import main as _incidence

from ..jacobians import _base
#]


class PeriodJacobian(_base.Jacobian, ):
    """
    """
    #[

    def eval(
        self,
        data_array: _np.ndarray,
        column_offset: int,
        steady_array: _np.ndarray | None,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(
            data_array, column_offset, steady_array,
        )
        return self._create_jacobian(self._shape, diff_array, self._map, )

    def _create_eid_to_wrts(
        self,
        equations: Iterable[_equations.Equation],
        all_wrt_qids: Iterable[int],
        /,
    ) -> dict[int, tuple[int, ...]]:
        """
        """
        return {
            eqn.id: tuple(
                qid for qid in all_wrt_qids
                if _incidence.is_qid_zero_in_tokens(eqn.incidence, qid, )
            )
            for eqn in equations
        }

    # ===== Implement AtomFactoryProtocol =====

    def create_data_index_for_token(
        self,
        token: _incidence.Token,
        /,
    ) -> tuple[int, slice]:
        """
        """
        return (token.qid, token.shift, )

    def create_diff_for_token(
        self,
        token: _incidence.Token,
        wrt_qids: tuple[int],
        /,
    ) -> _np.ndarray:
        """
        """
        if token is None:
            return _np.zeros((len(wrt_qids), 1, ), )
        if token.shift != 0:
            return 0
        try:
            index = wrt_qids.index(token.qid, )
            diff = _np.zeros((len(wrt_qids), 1, ), )
            diff[index] = 1
            return diff
        except:
            return 0

    #]




"""
Univariate time series filters
"""


#[
from __future__ import annotations

from numbers import (Real, )
from collections.abc import (Iterable, Callable, )
from types import (EllipsisType, )
from typing import (Self, )
import numpy as _np

from .. import pages as _pages
from .. import dates as _dates
from . import main as _series
from . import _functionalize
#]


__all__ = ("hpf", )


_AUTO_SMOOTH = {
    _dates.Frequency.YEARLY: (10*1)**2,
    _dates.Frequency.HALFYEARLY: (10*2)**2,
    _dates.Frequency.QUARTERLY: (10*4)**2,
    _dates.Frequency.MONTHLY: (10*12)**2,
    "default": (10*4)**2,
}


def _get_default_smooth(frequency, /, ):
    """
    Default smoothing parameter (lambda)
    """
    return _AUTO_SMOOTH.get(frequency, _AUTO_SMOOTH["default"])


class _ConstrainedHodrickPrescottFilter:
    """
    """
    #[
    def __init__(
        self,
        num_periods: int,
        smooth: Real,
        /,
        level_where: list[int] | None = None,
        change_where: list[int] | None = None,
        log: bool = False,
    ) -> None:
        self._num_periods = num_periods
        self._smooth = smooth
        self._log = log
        self._num_extra_rows = 0
        self._create_plain_filter_matrix()
        self._add_level_constraints(level_where, )
        self._add_change_constraints(change_where, )

    def filter_data(
        self,
        data: _np.ndarray,
        /,
        level_data: _np.ndarray | None = None,
        change_data: _np.ndarray | None = None,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        """
        """
        data = data.reshape(-1, 1)
        F = self._add_eye_for_observations(data, )
        extended_data = self._extend_data(data, level_data, change_data, )
        if self._log:
            extended_data = _np.log(extended_data, )
        enforced_zeros_where = _np.where(_np.isnan(data, ))
        extended_data[enforced_zeros_where] = 0
        trend_data = _np.linalg.solve(F, extended_data, )
        extended_data[enforced_zeros_where] = _np.nan
        gap_data = extended_data - trend_data
        if self._num_extra_rows > 0:
            trend_data = trend_data[:-self._num_extra_rows, :]
            gap_data = gap_data[:-self._num_extra_rows, :]
        if self._log:
            trend_data = _np.exp(trend_data, )
            gap_data = _np.exp(gap_data, )
        return trend_data, gap_data

    def _add_eye_for_observations(
        self,
        data: _np.ndarray,
        /,
    ) -> None:
        """
        Add 1 to the F diagonal for each row where an observation is available
        """
        #[
        F = _np.copy(self._F)
        quasi_eye = _np.diag(_np.float64(~_np.isnan(data.flatten(), )))
        F[:self._num_periods, :self._num_periods] += quasi_eye
        return F
        #]

    def _extend_data(
        self,
        data: _np.ndarray,
        level_data: _np.ndarray | None,
        change_data: _np.ndarray | None,
    ) -> _np.ndarray:
        """
        """
        extended_data = data.copy() if data.ndim > 1 else data.copy().reshape(-1, 1)
        tile_dim = (1, extended_data.shape[1], )
        if level_data is not None:
            extended_data = _np.vstack((extended_data, _np.tile(level_data, tile_dim), ), )
        if change_data is not None:
            extended_data = _np.vstack((extended_data, _np.tile(change_data, tile_dim), ), )
        return extended_data

    def _create_plain_filter_matrix(self, ) -> None:
        """
        """
        K = _np.zeros((self._num_periods-2, self._num_periods), dtype=float)
        for i in range(self._num_periods-2):
            K[i,i] = 1
            K[i,i+2] = 1
        for i in range(self._num_periods-2):
            K[i,i+1] = -2
        self._F = self._smooth * (K.T @ K)

    def _add_level_constraints(self, level_where: list[int], /, ):
        if not level_where:
            return
        num_constraints = len(level_where)
        self._num_extra_rows += num_constraints
        extra_rows = _np.zeros((num_constraints, self._num_periods, ), dtype=float)
        extra_variants = _np.zeros((self._num_periods + num_constraints, num_constraints, ), dtype=float)
        for i, j in enumerate(level_where, ):
            extra_rows[i, j] = 1
            extra_variants[j, i] = 1
        self._F = _np.vstack((self._F, extra_rows, ))
        self._F = _np.hstack((self._F, extra_variants, ))

    def _add_change_constraints(self, change_where: list[int], /, ):
        if not change_where:
            return
        num_constraints = len(change_where)
        extra_rows = _np.zeros((num_constraints, self._F.shape[1], ), dtype=float, )
        extra_variants = _np.zeros((self._F.shape[0] + num_constraints, num_constraints, ), dtype=float, )
        for i, j in enumerate(change_where, ):
            extra_rows[i, [j-1, j]] = (-1, 1)
            extra_variants[[j-1, j], i] = (-1, 1)
        self._F = _np.vstack((self._F, extra_rows, ))
        self._F = _np.hstack((self._F, extra_variants, ))
        self._num_extra_rows += num_constraints
    #]


class Inlay:
    """
    Constrained Hodrick-Prescott filter mixin for Series
    """
    #[

    @_pages.reference(
        category="filtering",
    )
    def hpf(self, /, ):
        r"""
················································································

==Constrained Hodrick-Prescott filter==


### Functional forms creating a new time `Series` object ###


```
trend, gap = irispie.hpf(
    self,
    /,
    span=None,
    smooth=None,
    log=False,
    level=None,
    change=None,
)
```


### Class methods changing an existing time `Series` object in-place ###


```
self.hpf_trend(
    /,
    span=None,
    smooth=None,
    log=False,
    level=None,
    change=None,
)
```

```
self.hpf_gap(
    /,
    span=None,
    smooth=None,
    log=False,
    level=None,
    change=None,
)
```


### Input arguments ###


???+ input "self"
    A time `Series` object whose data are filtered. All the
    observations contained in the input time series are used in the
    Hodrick-Prescott filter calculations no matter the time `span`
    specified; the time `span` only determines the time span of the output
    series.

???+ input "span"
    Time span on which the trend component of the Hodrick-Prescott filter
    is calculated. If `span=None`, the results are returned on
    the time span of the original series, `self`.

???+ input "smooth"
    Smoothing parameter (also known as $\lambda$) for the Hodrick-Prescott filter. If `smooth=None`,
    a default value is used based on the frequency of the input series `self`:

    | Date frequency | Default `smooth` ($\lambda$)
    |----------------|-----------------------------:
    | `YEARLY`       |                         100
    | `HALF-YEARLY`  |                         400
    | `QUARTERLY`    |                       1,600
    | `MONTHLY`      |                     144,000
    | Otherwise      |                       1,600

???+ input "log"
    If `log=True`, the Hodrick-Prescott filter is calculated on the logarithm
    of the input data, and the results are delogarithmized back to the original
    scale.

???+ input "level"
    A `Series` object with the level constraints for the Hodrick-Prescott
    filter (aka judgmental adjustments). If `level=None`, no level
    constraints are imposed. If `log=True`, the level constraints are
    logarithmized first before entering the calculations.

???+ input "change"
    A `Series` object with the change constraints for the Hodrick-Prescott
    filter (aka judgmental adjustments). If `change=None`, no change
    constraints are imposed. If `log=True`, the change constraints are
    logarithmized first before entering the calculations; this effectively
    means that for `log=True`, the `change` constraints need to be
    expressed as gross rates of change (period on period).


### Returns ###


???+ returns "trend"
    A new `Series` object with the trend component of the Hodrick-Prescott
    filter. The trend component is always returned on the entier time `span`
    specified regardless of the actual time span of the original series.

???+ returns "gap"
    A new `Series` object with the trend component of the Hodrick-Prescott
    filter. The gap component is calculated on the time `span` as the
    difference between the actual observations and the trend component;
    therefore, unline the `trend` series, the `gap` series is defined only
    in time periods where the actual observations are available in the
    original series.

???+ returns "self"
    The existing `Series` object with its values replaced in-place by the trend
    component of the Hodrick-Prescott filter.


### Details ###


???+ abstract "Time span of HP filter calculations"

    The Hodrick-Prescott filter is calculated on a time span starting in
    the period that is the first common period of

    * the input data,
    * the `span` argument,
    * the level constraints, and
    * the change constraints.

    and ending in the period that is the last common period of the same
    four data sources.

    The actual time series returned is then clipped to comply with the
    `span` argument if necessary.


???+ abstract "Math description"

    The constrained Hodrick-Prescott filter is a method for decomposing a
    time series into a lower-frequency (trend) component and a
    higher-frequency (cyclical) component subject to level and/or change
    constraints. The filter is implemented as the following constrained
    dynamic optimization problem

    $$
    \begin{gathered}
    \min\nolimits_{\{\overline{y}_t\}}
        \left[
            \lambda \sum_{t\in\Omega}
            \left( y_t - \overline{y}_t \right)^2
            + \sum_{t=3}^{T}
            \left( \Delta \overline{y}_t - \Delta \overline{y}_{t-1} \right)^2
        \right]
        \\[10pt]
    \text{subject to} \\[10pt]
    \overline{y}_t = L_t \ \forall t \in \Omega_L \\[10pt]
    \Delta \overline{y}_t = C_t \ \forall t \in \Omega_C
    \end{gathered}
    $$

    where

    * $y_t$ is the original time series data,
    * $\overline{y}_t$ is the calculated trend component,
    * $L_t$ is the level constraint data,
    * $C_t$ is the change constraint data,
    * $\lambda$ is the smoothing parameter,
    * $1, \dots, T$ is the HP filter time span (see above),
    * $\Omega$ is the time periods within the filter time span where the original data are available,
    * $\Omega_L$ is the time periods within the filter time span where the level constraint data are specified,
    * $\Omega_C$ is the time periods within the filter time span where the change constraint data are specified.

    The gap component, $\widehat{y}_t$, is then calculated as the difference between the
    original data and the trend component,

    $$
    \widehat{y}_t \equiv y_t - \overline{y}_t
    $$

················································································
        """
        raise NotImplementedError

    def hpf_trend(self, /, *args, **kwargs):
        """
        """
        start_date, trend_data, _ = _data_hpf(self, *args, **kwargs, )
        self._replace_start_date_and_values(start_date, trend_data, )

    def hpf_gap(self, /, *args, **kwargs):
        """
        """
        start_date, _, gap_data = _data_hpf(self, *args, **kwargs, )
        self._replace_start_date_and_values(start_date, gap_data, )

    #]


def _data_hpf(
    self,
    *,
    span: Iterable[_dates.Dater] | EllipsisType = ...,
    smooth: Real | None = None,
    log: bool = False,
    level: _series.Series | None = None,
    change: _series.Series | None = None,
) -> tuple[_dates.Dater, _np.ndarray, _np.ndarray]:
    """
    Hodrick-Prescott filter run on a multi-variant data matrix
    """
    #[
    if smooth is None:
        smooth = _get_default_smooth(self.frequency, )
    span = self._resolve_dates(span, )
    encompassing_span, *from_to = _dates.get_encompassing_span(self, level, change, span, )
    num_periods = len(encompassing_span, )
    level_data, level_where = _prepare_constraints(level, from_to, )
    change_data, change_where = _prepare_constraints(change, from_to, )
    change_data, change_where = _remove_first_date_change(change_data, change_where, )
    #
    hp = _ConstrainedHodrickPrescottFilter(
        num_periods,
        smooth,
        level_where=level_where,
        change_where=change_where,
        log=log,
    )
    #
    trend_data = []
    gap_data = []
    for data_variant in self.iter_own_data_variants_from_to(from_to, ):
        trend_data_variant, gap_data_variant = hp.filter_data(
            data_variant,
            level_data=level_data,
            change_data=change_data,
        )
        trend_data.append(trend_data_variant, )
        gap_data.append(gap_data_variant, )
    #
    trend_data = _np.hstack(trend_data, )
    gap_data = _np.hstack(gap_data, )
    #
    if span:
        new_start_date = min(span, )
        new_end_date = max(span, )
        clip_start = new_start_date - from_to[0]
        clip_end = new_end_date - from_to[0] + 1
        trend_data = trend_data[clip_start:clip_end, ...]
        gap_data = gap_data[clip_start:clip_end, ...]
    else:
        new_start_date = None
        trend_data = trend_data[[], ...]
        gap_data = gap_data[[], ...]
    #
    return new_start_date, trend_data, gap_data
    #]


def hpf(self, *args, **kwargs, ) -> tuple[_series.Series, _series.Series]:
    """
    Constrained Hodrick-Prescott filter
    """
    start_date, trend_data, gap_data = _data_hpf(self, *args, **kwargs, )
    trend = type(self)(start_date=start_date, values=trend_data, )
    gap = type(self)(start_date=start_date, values=gap_data, )
    return trend, gap


for n in ("hpf_trend", "hpf_gap", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )



def _prepare_constraints(
    constraint: Self | None,
    from_to: tuple[_dates.Dater, _dates.Dater, ],
    /,
) -> tuple[_np.ndarray | None, list[int] | None, ]:
    #[
    if constraint is None:
        return None, None
    data = constraint.get_data_from_to(from_to, 0, )
    where = list(_np.where(~_np.isnan(data.reshape(-1, ), ), )[0])
    data = data[where, :]
    return data, where
    #]


def _remove_first_date_change(
    change_data: _np.ndarray | None,
    change_where: list[int] | None,
) -> tuple[_np.ndarray, list[int], ]:
    #[
    if change_where and 0 in change_where:
        where = change_where.index(0)
        change_where.pop(where)
        change_data = _np.delete(change_data, where, axis=0)
        if not change_where:
            change_data = None
            change_where = None
    return change_data, change_where
    #]






import sys
sys.path.append("../../../src")
import irispie as _ip

db = _ip.Databank()
db.date = _ip.Dater.from_sdmx_string(_ip.Frequency.QUARTERLY, "2020-Q1")
db.date_string = db.date.to_sdmx_string()

assert db.date_string == "2020-Q1"
assert db.date == _ip.qq(2020,1)




"""
Simple temporary flyweight implementation
"""


#[
from __future__ import annotations
#]


class FlyweightRegister:
    """
    """
    #[

    def __init__(self, flyweight_factory: type, ) -> None:
        """
        """
        self._flyweight_factory = flyweight_factory
        self._flyweights = {}

    def __call__(self, *args, ):
        """
        """
        key = args
        if key not in self._flyweights:
            self._flyweights[key] = self._flyweight_factory(*args, )
        return self._flyweights[key]

    #]




"""
Documentation introspection tools
"""


#[
from __future__ import annotations

from collections.abc import (Callable, )
import re as _re
import os as _os
#]


_TAGLINE_PATTERN = _re.compile(r"==(.*?)==", )


def reference(**kwargs, ) -> Callable:
    def _decorate(callable_obj, ) -> Callable:
        callable_obj._pages_reference = True
        for k, v in kwargs.items():
            setattr(callable_obj, f"_pages_{k}", v)
        if not hasattr(callable_obj, "_pages_category", ):
            callable_obj._pages_category = None
        if not hasattr(callable_obj, "_pages_call_name", ):
            callable_obj._pages_call_name = f"{callable_obj.__name__}"
        if not hasattr(callable_obj, "_pages_priority", ):
            callable_obj._pages_priority = _PRIORITY.get(callable_obj._pages_category, _PRIORITY[None], )
        if not hasattr(callable_obj, "_pages_add_heading", ):
            callable_obj._pages_add_heading = type(callable_obj, ) is not type
        if not hasattr(callable_obj, "_pages_call_name_is_code", ):
            callable_obj._pages_call_name_is_code = True
        if callable_obj._pages_add_heading:
            callable_obj.__doc__ = _add_heading(
                callable_obj.__doc__,
                callable_obj._pages_call_name,
                callable_obj._pages_call_name_is_code,
            )
        callable_obj._pages_tagline = _extract_tagline(callable_obj, )

        return callable_obj
    return _decorate


def _extract_tagline(callable_obj, ) -> str:
    m = _TAGLINE_PATTERN.search(callable_obj.__doc__, )
    return m.group(1, ) if m else ""


def _add_heading(docstring: str, call_name: str, call_name_is_code: bool, ) -> str:
    icon = "☐"
    heading_text = f"{icon} `{call_name}`" if call_name_is_code else call_name
    heading_decor = "\n" + "-"*len(heading_text) + "--\n"
    docstring = _remove_visual_dividers(docstring, )
    docstring = "\n" + heading_text + heading_decor + docstring
    docstring = _add_visual_dividers(docstring, )
    return docstring


_PRIORITY = {
    "constructor": 10,
    None: 0,
}


_DIVIDER_PATTERN = _re.compile(r"\n·{20,}\n", )
_DIVIDER_PATTERN_LEGACY = _re.compile(r"\n\.{20,}\n", )


def _remove_visual_dividers(docstring: str, ) -> str:
    docstring = _DIVIDER_PATTERN.sub("", docstring)
    docstring = _DIVIDER_PATTERN_LEGACY.sub("", docstring)
    return docstring


def _add_visual_dividers(docstring: str, ) -> str:
    divider = "\n" + "·"*80 + "\n"
    return divider + docstring + divider





"""
"""


#[
from __future__ import annotations

from typing import (Protocol, Callable, )
from numbers import (Real, )
import numpy as _np

from ..incidences import main as _incidence
#]


_WHEN_DATA_SYMBOL = {
    True: "?",
    False: "!",
    None: "",
}


class PlanTransformProtocol(Protocol, ):
    """
    """
    #[

    _DEFAULT_NAME_FORMAT: str
    _SYMBOL: str

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray | None,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        ...

    def __str__(self, ) -> str: ...

    #]


class PlanTransform:
    """
    """
    #[

    def __init__(
        self,
        /,
        when_data: bool | None = False,
        name_format: str | None = None,
        shift: int = -1,
    ) -> None:
        self.when_data = when_data or False
        self._name_format = name_format or self._DEFAULT_NAME_FORMAT
        self._shift = shift

    @property
    def symbol(self, /, ) -> str:
        return _WHEN_DATA_SYMBOL[self.when_data] + self._SYMBOL

    def __str__(self, /, ) -> str:
        return self.symbol

    def __repr__(self, /, ) -> self:
        return self.__str__()

    def resolve_databox_name(
        self,
        base_name: str, 
        /,
    ) -> str | None:
        """
        """
        return (
            self._name_format.format(base_name, )
            if self._name_format is not None else None
        )

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray | None,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        ...

    #]


class PlanTransformNone(PlanTransform, ):
    """
    """
    #[
    _DEFAULT_NAME_FORMAT = "{}"
    _SYMBOL = "[=]"

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        return exogenized_values_after[0]
    #]


class PlanTransformLog(PlanTransform, ):
    """
    """
    #[
    _DEFAULT_NAME_FORMAT = "log_{}"
    _SYMBOL = "[log]"

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        return _np.exp(exogenized_values_after[0])
    #]


class PlanTransformDiff(PlanTransform, ):
    """
    """
    #[
    _DEFAULT_NAME_FORMAT = "diff_{}"
    _SYMBOL = "[diff]"

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        return values_before[self._shift] + exogenized_values_after[0]
    #]


class PlanTransformDiffLog(PlanTransform, ):
    """
    """
    #[
    _DEFAULT_NAME_FORMAT = "diff_log_{}"
    _SYMBOL = "[diff_log]"

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        return values_before[self._shift] * _np.exp(exogenized_values_after[0])
    #]


class PlanTransformRoc(PlanTransform, ):
    """
    """
    #[
    _DEFAULT_NAME_FORMAT = "roc_{}"
    _SYMBOL = "[roc]"

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray | None,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        return values_before[self._shift] * exogenized_values_after[0]
    #]


class PlanTransformPct(PlanTransform, ):
    """
    """
    #[

    _DEFAULT_NAME_FORMAT = "pct_{}"
    _SYMBOL = "[pct]"

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        return values_before[self._shift] * (1 + exogenized_values_after[0]/100)

    #]


class PlanTransformFlat(PlanTransform, ):
    """
    """
    #[

    _DEFAULT_NAME_FORMAT = None
    _SYMBOL = "[flat]"

    def eval_exogenized(
        self,
        exogenized_values_after: _np.ndarray,
        values_before: _np.ndarray,
        /,
    ) -> Real:
        return values_before[self._shift]

    #]


CHOOSE_TRANSFORM_CLASS = {
    None: PlanTransformNone,
    "level": PlanTransformNone,
    "none": PlanTransformNone,
    "log": PlanTransformLog,
    "diff": PlanTransformDiff,
    "diff_log": PlanTransformDiffLog,
    "difflog": PlanTransformDiffLog,
    "roc": PlanTransformRoc,
    "pct": PlanTransformPct,
    "flat": PlanTransformFlat,
}


def resolve_transform(transform, **kwargs, ) -> PlanTransformProtocol:
    """
    """
    #[
    if transform is None or isinstance(transform, str):
        return CHOOSE_TRANSFORM_CLASS[transform](**kwargs, )
    else:
        return transform
    #]




"""Plans for dynamic simulations and steady state calculations"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Self, Any, Protocol, NoReturn, )
from types import (EllipsisType, )
import warnings as _wa
import functools as _ft
import copy as _copy

from ..conveniences import copies as _copies
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import pages as _pages
from . import _pretty as _pretty
from . import _indexes as _indexes
from . import transforms as _transforms
#]


CHOOSE_TRANSFORM_CLASS = _transforms.CHOOSE_TRANSFORM_CLASS


__all__ = (
    "PlanSimulate",
    "PlanSteady",
    "Plan",
    "CHOOSE_TRANSFORM_CLASS",
)


class PlannableSimulateProtocol(Protocol, ):
    """
    """
    #[

    simulate_can_be_exogenized: Iterable[str] | None
    simulate_can_be_endogenized: Iterable[str] | None

    #]


class PlannableSteadyProtocol(Protocol, ):
    """
    """
    #[

    steady_can_be_exogenized: Iterable[str] | None
    steady_can_be_endogenized: Iterable[str] | None
    steady_can_be_fixed_level: Iterable[str] | None
    steady_can_be_fixed_change: Iterable[str] | None

    #]


@_pages.reference(
    path=("structural_models", "simulation_plans.md", ),
    categories={
        "constructor": "Creating new simulation plans",
        "property": None,
        "definition": "Defining exogenized and endogenized data points",
        "information": "Getting information about simulation plans",
    },
)
class PlanSimulate(
    _pretty.PrettyMixin,
    _indexes.ItemMixin,
    _copies.CopyMixin,
):
    """
················································································

Simulation plans
=================

`PlanSimulate` objects are used to set up conditioning assumptions
for simulating [`Simultaneous`](simultaneous_modelsd) or
[`Sequential`](sequential_models) models. The simulation plans specify

* what variables to exogenize in what periods
* what shocks to endogenized in what periods (`Simultaneous` models only)

The plans only contain meta information, not the actual data points for the
exogenized variables. The actual data points are included in the input databox.

················································································
    """
    #[

    _registers = (
        "exogenized",
        "endogenized",
    )

    __slots__ = (
        *tuple(f"can_be_{r}" for r in _registers),
        *tuple(f"_{r}_register" for r in _registers),
        *tuple(f"default_{r}" for r in _registers),
        "base_span",
    )

    def _properties():
        """
················································································


Directly accessible properties of `PlanSimulate` objects
=========================================================

Property | Description
---|---
`start_date` | Start date of the simulation span
`end_date` | End date of the simulation span
`num_periods` | Number of periods in the simulation span
`base_span` | Simulation span
`can_be_exogenized` | Names of quantities that can be exogenized
`can_be_endogenized` | Names of quantities that can be endogenized
`pretty` | Tabular view of the simulation plan

················································································
        """

    @_pages.reference(category="constructor", call_name="PlanSimulate", )
    def __init__(
        self,
        plannable: PlannableSimulateProtocol,
        span: Iterable[_dates.Dater] | None,
    ) -> None:
        """
················································································

==Create new simulation plan object==

```
self = PlanSimulate(model, time_span, )
```

Create a new simulation plan object for a
[`Simultaneous`](sequential_models) or
[`Sequential`](sequential_models) model.

### Input arguments ###

???+ input "model"

    A [`Simultaneous`](sequential_models) or
    [`Sequential`](sequential_models) model that will be simulated.

???+ input "time_span"

    A date range on which the `model` will be simulated.


### Returns ###

???+ returns "self"

    A new empty simulation plan object.

················································································
        """
        self.base_span = tuple(span)
        self._default_exogenized = None
        self._default_endogenized = None
        for r in self._registers:
            register = {
                n: [None] * self.num_periods
                for n in getattr(plannable, f"simulate_can_be_{r}")
            } if hasattr(plannable, f"simulate_can_be_{r}") else {}
            setattr(self, f"can_be_{r}", tuple(register.keys()))
            setattr(self, f"_{r}_register", register)

    def check_consistency(
        self,
        plannable: PlannableSimulateProtocol,
        span: Iterable[_dates.Dater] | None,
        /,
    ) -> None:
        """
        """
        benchmark = type(self)(plannable, span, )
        if self.base_span != benchmark.base_span:
            raise _wrongdoings.IrisPieError(f"Plan span must be the same as the simulation span")
        for r in self._registers:
            if getattr(self, f"can_be_{r}") != getattr(benchmark, f"can_be_{r}"):
                raise _wrongdoings.IrisPieError(f"Plan must be created using the simulated model")

    @property
    @_pages.reference(category="property", )
    def start_date(self, /, ) -> _dates.Dater:
        """==Start date of the simulation span=="""
        return self.base_span[0]

    @property
    @_pages.reference(category="property", )
    def end_date(self, /, ) -> _dates.Dater:
        """==End date of the simulation span=="""
        return self.base_span[-1]

    @property
    @_pages.reference(category="property", )
    def num_periods(self, /, ) -> int:
        """==Number of periods in the simulation span=="""
        return len(self.base_span) if self.base_span is not None else 1

    @property
    @_pages.reference(category="property", )
    def frequency(self, /, ) -> str:
        """==Date frequency of the simulation span=="""
        return self.start_date.frequency

    @_pages.reference(category="definition", )
    def exogenize(
        self,
        dates: Iterable[_dates.Dater] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        /,
        *,
        transform: str | None = None,
        # when_data: bool | None = None,
        **kwargs,
    ) -> None:
        """
················································································

==Exogenize certain quantities at certain dates==

```
self.exogenize(
    dates, names,
    /,
    transform=None,
    when_data=False,
)
```

### Input arguments ###


???+ input "dates"

    Dates at which the `names` will be exogenized; use `...` for all simulation dates.

???+ input "names"

    Names of quantities to exogenize at the `dates`; use `...` for all exogenizable quantities.


### Input arguments available only for `Sequential` models ###

???+ input "transform"

    Transformation (specified as a string) to be applied to the exogenized
    quantities; if `None`, no tranformation is applied.

???+ input "when_data"

    If `True`, the data point will be exogenized only if a proper value
    exists in the input data.

················································································
        """
        transform = _transforms.resolve_transform(transform, **kwargs, )
        self._plan_simulate(self._exogenized_register, dates, names, transform, )

    @_pages.reference(category="information", )
    def get_names_exogenized_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names exogenized at a certain date==

················································································
        """
        return self._get_names_registered_in_period(self._exogenized_register, *args, **kwargs, )

    @_pages.reference(category="definition", )
    def endogenize(
        self,
        dates: Iterable[_dates.Dater] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        /,
    ) -> None:
        """
················································································

==Endogenize certain quantities at certain dates==

················································································
        """
        transform = None
        new_status = True
        self._plan_simulate(self._endogenized_register, dates, names, new_status, )

    @_pages.reference(category="information", )
    def get_names_endogenized_in_period(self, *args, **kwargs, ) -> tuple[str, ...]:
        """
················································································

==Get names endogenized at a certain date==

················································································
        """
        return self._get_names_registered_in_period(self._endogenized_register, *args, **kwargs, )

    def swap(
        self,
        dates: Iterable[_dates.Dater] | EllipsisType,
        pairs: Iterable[tuple[str, str]] | tuple[str, str],
        *args, **kwargs,
    ) -> None:
        """
        """
        pairs = tuple(pairs)
        if not pairs:
            return
        if len(pairs) == 2 and isinstance(pairs[0], str) and isinstance(pairs[1], str):
            pairs = (pairs, )
        for pair in pairs:
            self.exogenize(dates, pair[0], *args, **kwargs, )
            self.endogenize(dates, pair[1], *args, **kwargs, )

    def get_exogenized_point(
        self,
        name: str,
        date: _dates.Dater,
        /,
    ) -> _transforms.PlanTransformProtocol | None:
        """
        """
        column = next((
            column
            for column, t in enumerate(self.base_span)
            if date == t
        ))
        point = self._exogenized_register[name][column]
        return point if point is not None else self._default_exogenized

    def get_endogenized_point(
        self,
        name: str,
        column: int,
        /,
    ) -> _transforms.PlanTransformProtocol | None:
        """
        """
        point = self._endogenized_register[name][column]
        return point if point is not None else self._default_endogenized

    def get_databox_names(self, /, ) -> tuple[str]:
        """
        """
        databox_names = set()
        for k, v in self._exogenized_register.items():
            databox_names.update(
                t.resolve_databox_name(k, )
                for t in v if t is not None
            )
        return tuple(n for n in databox_names if n is not None)

    def _get_names_registered_in_period(
        self,
        register: dict[str, Any],
        date: _dates.Dater,
    ) -> tuple[str, ...]:
        """
        """
        column_index = self.base_span.index(date)
        return tuple(
            name
            for name, status in register.items()
            if bool(status[column_index])
        )

    def __str__(self, /, ) -> str:
        """
        """
        return self.get_pretty_string()

    def _plan_simulate(
        self,
        register: dict,
        dates: Iterable[_dates.Dater] | EllipsisType,
        names: Iterable[str] | str | EllipsisType,
        new_status: Any,
    ) -> None:
        """
        """
        names = _resolve_and_check_names(register, names, )
        date_indices = self._get_date_indices(dates, )
        for n in names:
            for t in date_indices:
                register[n][t] = new_status

    def _get_date_indices(
        self,
        dates: Iterable[_dates.Dater] | EllipsisType,
        /,
    ) -> tuple[int, ...]:
        """
        """
        if dates is Ellipsis:
            return tuple(range(len(self.base_span)))
        dates = dates.resolve(self, ) if hasattr(dates, "resolve") else dates
        catch_invalid_dates(dates, self.base_span, )
        return tuple(d - self.start_date for d in dates)

    #]


Plan = PlanSimulate


@_pages.reference(
    path=("structural_models", "steady_plans.md", ),
    categories={
        "constructor": "Creating new steady plans",
        "property": None,
        "definition": "Defining exogenized, endogenized and fixed quantities",
    },
)
class PlanSteady:
    """
················································································

Steady plans
=============

`PlanSteady` objects are used to define assumptions about the steady state
values of certain model variables.

················································································
    """
    #[

    __slots__ = (
        "can_be_exogenized",
        "can_be_endogenized",
        "can_be_fixed_level",
        "can_be_fixed_change",
        "_exogenized_register",
        "_endogenized_register",
        "_fixed_level_register",
        "_fixed_change_register",
    )

    @_pages.reference(category="constructor", call_name="PlanSteady", )
    def __init__(
        self,
        plannable: PlannableSteadyProtocol,
        /,
    ) -> None:
        """
················································································

==Create new steady plan object==

················································································
        """
        for r in ("exogenized", "endogenized", "fixed_level", "fixed_change", ):
            register = {
                n: False
                for n in getattr(plannable, f"steady_can_be_{r}")
            } if hasattr(plannable, f"steady_can_be_{r}") else {}
            setattr(self, f"can_be_{r}", tuple(register.keys()))
            setattr(self, f"_{r}_register", register)

    #]

#         self.exogenized = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_exogenized
#         } if plannable.can_be_exogenized else {}
#         #
#         self.endogenized = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_endogenized
#         } if plannable.can_be_endogenized else {}
#         #
#         self.fixed_level = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_fixed_level
#         } if plannable.can_be_fixed_level else {}
#         #
#         self.fixed_change = {
#             n: [None] * self.num_periods
#             for n in plannable.can_be_fixed_change
#         } if plannable.can_be_fixed_change else {}

    @_pages.reference(category="definition", )
    def exogenize(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Exogenize steady levels of variables==

················································································
        """
        pass

    @_pages.reference(category="definition", )
    def endogenize(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Endogenize parameters==

················································································
        """
        pass

    @_pages.reference(category="definition", )
    def fix_level(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Fix steady levels of variables==

················································································
        """
        pass

    @_pages.reference(category="definition", )
    def fix_change(
        self,
        names: Iterable[str] | str,
        /,
    ) -> None:
        """
················································································

==Fix steady changes of variables==

················································································
        """
        pass


def _resolve_and_check_names(
    register: dict | None,
    names: Iterable[str] | str | EllipsisType,
    /,
) -> tuple[str]:
    """
    """
    keys = tuple(register.keys()) if register else ()
    if names is Ellipsis:
        return keys
    if isinstance(names, str):
        names = (names, )
    names = tuple(names)
    invalid = [n for n in names if n not in keys]
    if invalid:
        raise _wrongdoings.IrisPieError(
            [f"Invalid names:"] + invalid
        )
    return names


def catch_invalid_dates(
    dates: Iterable[_dates.Dater],
    base_span: tuple[_dates.Dater],
    /,
) -> NoReturn | None:
    """
    """
    invalid = [repr(d) for d in dates if d not in base_span]
    if invalid:
        raise _wrongdoings.IrisPieError(
            ["These date(s) are out of simulation span:"] + invalid
        )




"""
"""


#[
from __future__ import annotations

from typing import (Any, Self, Iterable, )
import numpy as _np

from .. import dates as _dates
from ..databoxes import main as _databoxes
from . import _invariants as _invariants
#]


class Variant:
    """
    """
    #[

    __slots__ = (
        "data",
        "tag_alongs",
    )

    def __init__(
        self,
        /,
        tag_alongs: dict[str, Any] | None = None,
    ) -> None:
        """
        """
        self.data = None
        self.tag_alongs = tag_alongs or {}

    @classmethod
    def from_databox_variant(
        klass,
        databox_v: _databoxes.Databox | dict,
        invariant: _invariants.Invariant,
        /,
        fallbacks: dict[str, Number] | None = None,
        overwrites: dict[str, Number] | None = None,
    ) -> None:
        """
        """
        def _create_nan_vector() -> _np.ndarray:
            return _np.full((invariant.num_periods, ), _np.nan, dtype=_np.float64, )
        self = klass()
        data_list = []
        for n in invariant.names:
            new_data = _create_nan_vector()
            if n in databox_v:
                new_data[:] = databox_v[n]
            data_list.append(new_data, )
        self.data = _np.vstack(data_list, )
        self._fill_fallbacks(fallbacks, invariant, )
        self._fill_overwrites(overwrites, invariant, )
        return self

    @classmethod
    def nan_data_array(
        klass,
        invariant: _invariants.Invariant,
        /,
    ) -> Self:
        """
        """
        self = klass()
        self.data = _np.full((invariant.num_names, invariant.num_periods, ), _np.nan, dtype=_np.float64, )
        return self

    def copy(self, /) -> Self:
        """
        """
        return type(self).from_data_array(self.data.copy(), )

    def remove_periods_from_start(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.data = self.data[:, remove:]

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.data = self.data[:, :-remove]

    def retrieve_record(
        self,
        record_id: int,
        /,
        columns: slice | Iterable[int] | None = None,
    ) -> _np.ndarray:
        """
        """
        columns = columns if columns is not None else ...
        return self.data[record_id, columns]

    def iter_data(self, /, ) -> _np.ndarray:
        """
        """
        return iter(self.data)

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        /,
        columns: slice | Iterable[int] | None = None,
    ) -> None:
        """
        """
        if columns is None:
            self.data[record_id, :] = values
        else:
            self.data[record_id, columns] = values

    def _fill_fallbacks(
        self,
        fallbacks: dict[str, Number] | None,
        invariant: _invariants.Invariant,
        /,
    ) -> None:
        """
        """
        if not fallbacks:
            return
        for record_id, name in enumerate(invariant.names, ):
            if name in fallbacks:
                values = self.retrieve_record(record_id, )
                index_nan = _np.isnan(values)
                values[index_nan] = _np.float64(fallbacks[name])
                self.store_record(values, record_id, )

    def _fill_overwrites(
        self,
        overwrites: dict[str, Number] | None,
        invariant: _invariants.Invariant,
        /,
    ) -> None:
        """
        """
        if not overwrites:
            return
        for record_id, name in enumerate(invariant.names, ):
            if name in overwrites:
                values = self.retrieve_record(record_id, )
                values[:] = _np.float64(overwrites[name])
                self.store_record(values, record_id, )




"""
"""


#[
from __future__ import annotations

from typing import (Callable, )
from numbers import (Real, )
import dataclasses as _dc
import copy as _cp
import numpy as _np

from .. import dates as _dates
from .. import has_variants as _has_variants
from ..conveniences import descriptions as _descriptions
from ..series import main as _series
from ..databoxes import main as _databoxes
#]


#
# REFACTOR:
# * rename `index_base_columns` to `are_base_columns` or use int indexes
#


class Invariant:
    """
    """
    #[

    __slots__ = (
        "descriptions",
        "logly_indexes",
        "index_base_columns",
        "names",
        "output_qids",
        "dates",
        "min_max_shift",
    )

    def __init__(
        self,
        names: Iterable[str],
        dates: Iterable[_dates.Dater],
        /,
        *,
        base_columns: Iterable[int] | None = None,
        descriptions: Iterable[str | None] | None = None,
        output_names: Iterable[str] | None = None,
        qid_to_logly: dict[int, bool | None] | None = None,
        min_max_shift: tuple[int, int] = (0, 0),
        frequency: _dates.Frequency | None = None,
        tag_alongs: dict[str, Any] | None = None,
    ) -> None:
        """
        """
        self.names = tuple(names)
        self.dates = tuple(dates)
        base_columns = base_columns or ()
        self.index_base_columns = tuple(i in base_columns for i in range(self.num_periods))
        self._populate_descriptions(descriptions, )
        self._populate_logly_index(qid_to_logly, )
        self.min_max_shift = min_max_shift
        self._populate_output_qids(output_names, )

    def _populate_output_qids(
        self,
        output_names: Iterable[str] | None,
    ) -> None:
        """
        """
        output_names = (
            tuple(set(output_names) & set(self.names))
            if output_names is not None
            else self.names
        )
        self.output_qids = tuple(
            qid
            for qid, name in enumerate(self.names, )
            if name in output_names
        )

    @property
    def num_periods(self, /, ) -> int:
        """
        """
        return len(self.dates)

    @property
    def num_names(self, /, ) -> int:
        """
        """
        return len(self.names)

    @property
    def from_to(self, /, ) -> tuple[_dates.Dater, _dates.Dater]:
        """
        """
        return self.dates[0], self.dates[-1]

    @property
    def base_columns(self, /, ) -> tuple[int]:
        """
        """
        return tuple(i for i in range(self.num_periods) if self.index_base_columns[i])

    @property
    def nonbase_columns(self, /, ) -> tuple[int]:
        """
        """
        return tuple(i for i in range(self.num_periods) if not self.index_base_columns[i])

    @property
    def base_slice(self, /, ) -> slice:
        """
        """
        base_columns = self.base_columns
        return slice(base_columns[0], base_columns[-1]+1)

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return {
            name: row
            for row, name in enumerate(self.names, )
        }

    def _populate_descriptions(
        self,
        descriptions: Iterable[str | None] | None,
    ) -> None:
        """
        """
        if descriptions is None:
            self.descriptions = tuple("" for _ in self.names)
        else:
            self.descriptions = tuple((d or "") for d in descriptions)

    def _populate_logly_index(
        self,
        qid_to_logly: dict[int, bool] | None,
        /,
    ) -> None:
        """
        """
        qid_to_logly = qid_to_logly or {}
        self.logly_indexes = tuple(
            i for i in range(len(self.names), )
            if qid_to_logly.get(i, False)
        )

    def remove_periods_from_start(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.dates = self.dates[remove:]
            self.index_base_columns = self.index_base_columns[remove:]

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove:
            self.dates = self.dates[:-remove]
            self.index_base_columns = self.index_base_columns[:-remove]





"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np
import wlogging as _wl

from .. import quantities as _quantities
from ..databoxes import main as _databoxes
from ..fords import simulators as _ford_simulators
from ..fords import solutions as _solutions
from ..periods import simulators as _period_simulators
from ..plans import main as _plans
from ..dataslates import main as _dataslates

from . import main as _simultaneous
#]


_dataslate_constructor = _dataslates.Dataslate.from_databox_for_slatable


class SimulateMixin:
    """
    """
    #[

    def simulate(
        self,
        input_db: _databoxes.Databox,
        span: Iterable[Dater],
        /,
        *,
        plan: _plans.PlanSimulate | None = None,
        method: Literal["first_order", "period", "stacked"] = "first_order",
        prepend_input: bool = True,
        target_databox: _databoxes.Databox | None = None,
        num_variants: int | None = None,
        remove_initial: bool = True,
        remove_terminal: bool = True,
        shocks_from_databox: bool = True,
        logging_level: int = _wl.INFO,
        **kwargs,
    ) -> tuple[_databoxes.Databox, dict[str, Any]]:
        """
        """
        logger = _wl.get_colored_logger(__name__, level=logging_level, )
        num_variants = self.num_variants if num_variants is None else num_variants
        base_dates = tuple(span, )
        #
        work_db = input_db.shallow()
        if not shocks_from_databox:
            shock_names = self.get_names(kind=_quantities.ANY_SHOCK, )
            work_db.remove(shock_names, strict=False, )
        #
        extra_databox_names = None
        if plan is not None:
            plan.check_consistency(self, base_dates, )
            extra_databox_names = plan.get_databox_names()
        #
        dataslate = _dataslate_constructor(
            self, work_db, base_dates,
            num_variants=num_variants,
            extra_databox_names=extra_databox_names,
        )
        #
        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            dataslate.iter_variants(),
        )
        #
        #=======================================================================
        # Main loop over variants
        for vid, model_v, dataslate_v in zipped:
            #
            # Simulate and write to dataslate
            _SIMULATE_METHODS[method](
                model_v, dataslate_v, vid,
                plan=plan,
                **kwargs,
            )
        #=======================================================================
        #
        # Remove initial and terminal condition data (all lags and leads
        # before and after the simulation span)
        if remove_terminal:
            dataslate.remove_terminal()
        if remove_initial:
            dataslate.remove_initial()
        #
        # Convert all variants of the dataslate to a databox
        output_db = dataslate.to_databox()
        if prepend_input:
            output_db.prepend(input_db, base_dates[0]-1, )
        #
        # Add to custom databox
        if target_databox is not None:
            output_db = target_databox | output_db
        #
        # Simulation info
        info = {}
        #
        return output_db, info

    #]


def _simulate_first_order(
    model: _simultaneous.Simultaneous,
    dataslate: _dataslates.Dataslate,
    vid: int,
    **kwargs,
) -> None:
    """
    """
    _ford_simulators.simulate_flat(
        model._variants[0].solution,
        model._solution_vectors,
        dataslate,
        vid,
        **kwargs,
    )


_SIMULATE_METHODS = {
    "first_order": _simulate_first_order,
    "period": _period_simulators.simulate,
}






import sys
sys.path.append("../../../src")

import irispie as _ip

y = _ip.Frequency.from_letter("y")
assert y is _ip.Frequency.YEARLY

h = _ip.Frequency.from_letter("h")
assert h is _ip.Frequency.HALFYEARLY

q = _ip.Frequency.from_letter("q")
assert q is _ip.Frequency.QUARTERLY

m = _ip.Frequency.from_letter("m")
assert m is _ip.Frequency.MONTHLY

d = _ip.Frequency.from_letter("d")
assert d is _ip.Frequency.DAILY

i = _ip.Frequency.from_letter("i")
assert i is _ip.Frequency.INTEGER

